using System;
using System.IO.Ports;
using System.Collections;
using PeteBrown.NetduinoMidiTest.Messages;

namespace PeteBrown.NetduinoMidiTest
{
    class MidiMessageReceiver : IDisposable
    {
        private SerialPort _midiIn =
            new SerialPort("COM1", 31250, Parity.None, 8, StopBits.One);

        private Queue _dataBuffer = new Queue();

        public event MidiMessageReceivedEventHandler MessageReceived;

        public void Open()
        {
            _midiIn.DataReceived += new SerialDataReceivedEventHandler(OnDataReceived);
            _midiIn.ErrorReceived +=new SerialErrorReceivedEventHandler(OnErrorReceived);
            _midiIn.Open();
        }

        public void Close()
        {
            if (_midiIn.IsOpen)
            {
                _midiIn.Close();
                _midiIn.DataReceived -= OnDataReceived;
                _midiIn.ErrorReceived -= OnErrorReceived;
            }
        }

        public void Dispose()
        {
            Close();
        }



        private void OnErrorReceived(object sender, SerialErrorReceivedEventArgs e)
        {

        }

        private void OnDataReceived(object sender, SerialDataReceivedEventArgs e)
        {

            // buffer for the data. May be more efficient to
            // preallocate a larger fixed-size buffer and 
            // read bytes into that (memory vs. cpu)
            byte[] readBuffer = new byte[_midiIn.BytesToRead];

            // Read the bytes from the buffer. There's no way
            // to read bytes without allocating a buffer, so
            // we'll allocate a temp buffer before putting the
            // results in the queue
            _midiIn.Read(readBuffer, 0, readBuffer.Length);


            foreach (byte b in readBuffer)
            {
                _dataBuffer.Enqueue(b);
            }

            ParseMidiStream();
        }

        // reference: http://www.tonalsoft.com/pub/pitch-bend/pitch.2005-08-24.17-00.aspx
        // Channel messages are the primary messages used for controlling synthesizers
        // and for receiving input from MIDI controllers. Channel messages require two
        // or three bytes, depending on the specific message. The first byte is always 
        // divided into two nibbles (4-bits). The first nibble contains the message 
        // number, and the second nibble contains the channel number. Channels have a 
        // value between 0 and 15.
        //
        // reference: http://www.cs.cf.ac.uk/Dave/Multimedia/node158.html
        // - MIDI message includes a status byte and up to two data bytes. 
        // - Status byte 
        //   - The most significant bit of status byte is set to 1. 
        //   - The 4 low-order bits identify which channel it belongs to (four bits produce 16 possible channels). 
        //   - The 3 remaining bits identify the message. 
        // - The most significant bit of data byte is set to 0. 
        private bool DequeueCompleteMessage(ref MidiMessage message)
        {
            // peek lets you take a look without removing from queue
            byte firstByte = (byte)_dataBuffer.Peek(); 

            // parse the message and channel number
            int messageNumber = firstByte >> 4;
            int channelNumber = firstByte & (byte)0x0F;

            int expectedDataByteCount = 0;

            switch ((MidiMessageNumber)messageNumber)
            {
                case MidiMessageNumber.NoteOff:
                case MidiMessageNumber.NoteOn:
                case MidiMessageNumber.PolyphonicAftertouch:
                case MidiMessageNumber.ControlChange:
                case MidiMessageNumber.PitchWheelChange:
                    expectedDataByteCount = 2;
                    break;

                case MidiMessageNumber.ProgramChange:
                case MidiMessageNumber.ChannelPressureAftertouch:
                    expectedDataByteCount = 1;
                    break;
            }
            
            if (_dataBuffer.Count >= expectedDataByteCount + 1)
            {
                _dataBuffer.Dequeue();  // remove the message/channel

                message.MessageNumber = (MidiMessageNumber)messageNumber;
                message.Channel = (MidiChannel)channelNumber;

                if (expectedDataByteCount >= 1)
                    message.DataByte1 = (byte)_dataBuffer.Dequeue();

                if (expectedDataByteCount >= 2)
                {
                    message.DataByte2 = (byte)_dataBuffer.Dequeue();
                    message.IsDataByte2Used = true;
                }

                // special case the pitch bend which stores a 16 bit value
                if (message.MessageNumber == MidiMessageNumber.PitchWheelChange)
                {
                    message.IsCombinedData = true;
                    message.DataBytesCombined = ((int)message.DataByte1 << 8) | (message.DataByte2);
                }

                return true;
            }
            else
            {
                return false;
            }
        }




        private void ParseMidiStream()
        {
            bool moreData = _dataBuffer.Count != 0;

            while (moreData)
            {
                MidiMessage message = new MidiMessage();

                if (DequeueCompleteMessage(ref message))
                {
                    if (MessageReceived != null)
                        MessageReceived(this,
                            new MidiMessageReceivedEventArgs() { Message = message });

                    // potentially more actionable data
                    moreData = _dataBuffer.Count != 0;
                }
                else
                {
                    // no more actionable data
                    moreData = false;
                }
            }
        }
    }
}
